package com.tzld.piaoquan.recommend.model

import ml.dmlc.xgboost4j.scala.spark.XGBoostClassifier
import org.apache.commons.lang.math.NumberUtils
import org.apache.commons.lang3.StringUtils
import org.apache.hadoop.io.compress.GzipCodec
import org.apache.spark.ml.evaluation.BinaryClassificationEvaluator
import org.apache.spark.ml.feature.VectorAssembler
import org.apache.spark.rdd.RDD
import org.apache.spark.sql.types.DataTypes
import org.apache.spark.sql.{Dataset, Row, SparkSession}

import java.util
import scala.io.Source
import scala.util.Random

object train_recsys_61_xgb_rov_20241209 {
  def main(args: Array[String]): Unit = {
    val spark = SparkSession
      .builder()
      .appName(this.getClass.getName)
      .getOrCreate()
    val sc = spark.sparkContext

    val param = ParamUtils.parseArgs(args)
    val featureFile = param.getOrElse("featureFile", "20241209_recsys_rov_name.txt")
    val trainPath = param.getOrElse("trainPath", "/dw/recommend/model/61_recsys_rov_train_data/20241210")
    val testPath = param.getOrElse("testPath", "")
    val savePath = param.getOrElse("savePath", "/dw/recommend/model/61_recsys_rov_predict_data/")
    val featureFilter = param.getOrElse("featureFilter", "XXXXXX").split(",")
    val eta = param.getOrElse("eta", "0.01").toDouble
    val gamma = param.getOrElse("gamma", "0.0").toDouble
    val max_depth = param.getOrElse("max_depth", "5").toInt
    val num_round = param.getOrElse("num_round", "100").toInt
    val num_worker = param.getOrElse("num_worker", "20").toInt
    val func_object = param.getOrElse("func_object", "binary:logistic")
    val func_metric = param.getOrElse("func_metric", "auc")
    val repartition = param.getOrElse("repartition", "20").toInt
    val negRate = param.getOrElse("negRate", "1.0").toDouble
    val modelPath = param.getOrElse("modelPath", "/dw/recommend/model/61_recsys_rov_model/model_xgb")
    val modelFile = param.getOrElse("modelFile", "model_xgb_for_recsys_rov.tar.gz")

    val loader = getClass.getClassLoader
    val resourceUrl = loader.getResource(featureFile)
    val content =
      if (resourceUrl != null) {
        val content = Source.fromURL(resourceUrl).getLines().mkString("\n")
        Source.fromURL(resourceUrl).close()
        content
      } else {
        ""
      }
    println(content)

    val features = content.split("\n")
      .map(r => r.replace(" ", "").replaceAll("\n", ""))
      .filter(r => r.nonEmpty || !featureFilter.contains(r))
    println("features.size=" + features.length)

    val trainData = createData(
      negRate,
      sc.textFile(trainPath),
      features
    )
    println("recsys rov:train data size：" + trainData.count())

    val fields = Array(
      DataTypes.createStructField("label", DataTypes.IntegerType, true)
    ) ++ features.map(f => DataTypes.createStructField(f, DataTypes.DoubleType, true))

    val schema = DataTypes.createStructType(fields)
    val trainDataSet: Dataset[Row] = spark.createDataFrame(trainData, schema)
    val vectorAssembler = new VectorAssembler().setInputCols(features).setOutputCol("features")
    val xgbInput = vectorAssembler.transform(trainDataSet).select("features", "label")
    //    val xgbParam = Map("eta" -> 0.01f,
    //      "max_depth" -> 5,
    //      "objective" -> "binary:logistic",
    //      "num_class" -> 3)
    val xgbClassifier = new XGBoostClassifier()
      .setEta(eta)
      .setGamma(gamma)
      .setMissing(0.0f)
      .setMaxDepth(max_depth)
      .setNumRound(num_round)
      .setSubsample(0.8)
      .setColsampleBytree(0.8)
      .setScalePosWeight(1)
      .setObjective(func_object)
      .setEvalMetric(func_metric)
      .setFeaturesCol("features")
      .setLabelCol("label")
      .setNthread(1)
      .setNumWorkers(num_worker)
      .setSeed(2024)
      .setMinChildWeight(1)
    val model = xgbClassifier.fit(xgbInput)

    if (modelPath.nonEmpty && modelFile.nonEmpty) {
      model.write.overwrite.save(modelPath)
      //      val gzPath = modelPath + "/" + modelFile
      //      CompressUtil.compressDirectoryToGzip(modelPath, gzPath)
    }

    if (testPath.nonEmpty) {
      val testData = createData(
        1.0,
        sc.textFile(testPath),
        features
      )
      val testDataSet = spark.createDataFrame(testData, schema)
      val testDataSetTrans = vectorAssembler.transform(testDataSet).select("features", "label")
      val predictions = model.transform(testDataSetTrans)

      println("recsys rov:columns:" + predictions.columns.mkString(",")) //[label, features, probability, prediction, rawPrediction]
      val saveData = predictions.select("label", "rawPrediction", "probability").rdd
        .map(r => {
          (r.get(0), r.get(1), r.get(2)).productIterator.mkString("\t")
        })
      val hdfsPath = savePath
      if (hdfsPath.nonEmpty && hdfsPath.startsWith("/dw/recommend/model/")) {
        println("删除路径并开始数据写入:" + hdfsPath)
        MyHdfsUtils.delete_hdfs_path(hdfsPath)
        saveData.repartition(repartition).saveAsTextFile(hdfsPath, classOf[GzipCodec])
      } else {
        println("路径不合法，无法写入:" + hdfsPath)
      }

      val evaluator = new BinaryClassificationEvaluator()
        .setLabelCol("label")
        .setRawPredictionCol("probability")
        .setMetricName("areaUnderROC")
      val auc = evaluator.evaluate(predictions.select("label", "probability"))
      println("recsys rov:auc:" + auc)
    }
  }

  def createData(negRate: Double, data: RDD[String], features: Array[String]): RDD[Row] = {
    data.filter(r => {
        val line: Array[String] = StringUtils.split(r, '\t')
        val label: Int = NumberUtils.toInt(line(1))
        line.length > 10 && (label > 0 || new Random().nextDouble() <= negRate)
      })
      .map(r => {
        val line: Array[String] = StringUtils.split(r, '\t')
        // val logKey = line(0)
        val label: Int = NumberUtils.toInt(line(1))
        // val scoresMap = line(2)
        val map: util.Map[String, Double] = new util.HashMap[String, Double]
        for (i <- 3 until line.length) {
          val fv: Array[String] = StringUtils.split(line(i), ':')
          map.put(fv(0), NumberUtils.toDouble(fv(1), 0.0))
        }

        val v: Array[Any] = new Array[Any](features.length + 1)
        v(0) = label
        for (i <- 0 until features.length) {
          v(i + 1) = map.getOrDefault(features(i), 0.0d)
        }
        Row(v: _*)
      })
  }
}
